using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace ID3TagEditLite {
    public partial class FormEditDlg : Form {
        TTagRecord m_record;
        public Point m_localLocation;
        public Point m_parentDesktopLocation;
        bool m_bMulti = false;
        public string m_filePath = "";
        public string m_lastModControlName = "";

        public FormEditDlg() {
            InitializeComponent();
            cbxGenre.BeginUpdate();
            foreach (TGenreItem g in Program.theGenre.m_genres.Values) {
                cbxGenre.Items.Add(g.GetNameForMenu());
            }
            cbxGenre.EndUpdate();
        }

        public void Setup(TTagRecord record, string filePath, bool bMulti) {
            m_record = record;
            m_filePath = filePath;
            m_bMulti = bMulti;
            string lastModControlName = m_lastModControlName;

            cbxSaveEnc.Items.AddRange(TTagRecord.GetEncodingNames());
            cbxVersion.Items.AddRange(TTagRecord.GetVersionNames());

            if (m_bMulti) {
                edtFileName.Enabled = false;
                edtTitle.Enabled = false;
                edtComment.Enabled = false;
                edtTrack.Enabled = false;
            }
            else {
                edtFilePath.Text = m_filePath;
                edtFileName.Text = m_record.m_fileName;
                edtTitle.Text = m_record.m_title.m_str;
                edtComment.Text = m_record.m_comment.m_str;
                edtTrack.Text = m_record.GetTrackString();
            }

            edtAlbum.Text = m_record.m_album.m_str;
            edtArtist.Text = m_record.m_artist.m_str;
            edtTrackMax.Text = m_record.GetTrackMaxString();
            edtYear.Text = m_record.m_year;
            edtGenre.Text = m_record.GetGenreNumberString();

            TGenreItem g = Program.theGenre.GetGenreItemByGenreNumber(m_record.m_genre);
            if (g != null) {
                cbxGenre.SelectedIndex = Program.theGenre.m_genres.Values.IndexOf(g);
            }
            edtLoadEnc.Text = TTagRecord.GetEncodingString(m_record.m_loadEnc);
            cbxSaveEnc.SelectedIndex = (int)m_record.m_saveEnc;
            cbxVersion.SelectedIndex = (int)m_record.m_version;

            m_lastModControlName = lastModControlName;
        }

        private void FormEditDlg_Shown(object sender, EventArgs e) {
            if (m_lastModControlName.Length > 0) {
                Control[] c = this.Controls.Find(m_lastModControlName, true);
                if (c != null && c.Length > 0)
                    c[0].Focus();
            }
        }

        public TTagRecord GetTagRecord() {
            return m_record;
        }

        private void btnOK_Click(object sender, EventArgs e) {
            if (m_bMulti) {
            }
            else {
                m_record.SetPropertyOfFileName(edtFileName.Text);
                m_record.SetPropertyOfTitle(edtTitle.Text);
                m_record.SetPropertyOfComment(edtComment.Text);
                m_record.SetPropertyOfTrack(edtTrack.Text);
            }
            m_record.SetPropertyOfAlbum(edtAlbum.Text);
            m_record.SetPropertyOfArtist(edtArtist.Text);
            m_record.SetPropertyOfGenre(edtGenre.Text);
            m_record.SetPropertyOfTrackMax(edtTrackMax.Text);
            m_record.SetPropertyOfYear(edtYear.Text);
            m_record.SetPropertyOfSaveEnc((TTagEncoding)cbxSaveEnc.SelectedIndex);
            m_record.SetPropertyOfVersion((TTagVersion)cbxVersion.SelectedIndex);

            this.DialogResult = DialogResult.OK;
            this.Close();
        }

        private void cbxGenre_SelectedValueChanged(object sender, EventArgs e) {
            int index = cbxGenre.SelectedIndex;
            if (index < 0)
                return;
            edtGenre.Text = Program.theGenre.m_genres.Values[index].m_number.ToString();
        }

        private void edtGenre_TextChanged(object sender, EventArgs e) {
            if (edtGenre.Text.Length > 0) {
                int n = 0;
                if (TTools.StringToInteger(ref n, edtGenre.Text) && n >= 0) {
                    TGenreItem g = Program.theGenre.GetGenreItemByGenreNumber(n);
                    if (g != null) {
                        cbxGenre.SelectedIndex = Program.theGenre.m_genres.Values.IndexOf(g);
                    }
                }
            }
            m_lastModControlName = edtGenre.Name;
        }

        private void EditDlg_Load(object sender, EventArgs e) {
            Point p = m_parentDesktopLocation;
            p.X += m_localLocation.X;
            p.Y += m_localLocation.Y;
            this.DesktopLocation = p;
        }

        private void EditDlg_Move(object sender, EventArgs e) {
        }

        private void EditDlg_FormClosed(object sender, FormClosedEventArgs e) {
            m_localLocation.X = this.DesktopLocation.X - m_parentDesktopLocation.X;
            m_localLocation.Y = this.DesktopLocation.Y - m_parentDesktopLocation.Y;
        }

        private void cbxSaveEnc_SelectedIndexChanged(object sender, EventArgs e) {
            if ((TTagEncoding)cbxSaveEnc.SelectedIndex == TTagEncoding.UTF16) {
                if ((TTagVersion)cbxVersion.SelectedIndex < TTagVersion.V_2_2) {
                    cbxVersion.SelectedIndex = (int)TTagVersion.V_2_2;
                }
            }
            m_lastModControlName = cbxSaveEnc.Name;
        }

        private void cbxVersion_SelectedIndexChanged(object sender, EventArgs e) {
            if ((TTagVersion)cbxVersion.SelectedIndex < TTagVersion.V_2_2) {
                cbxSaveEnc.SelectedIndex = (int)TTagEncoding.ANSI;
            }
            m_lastModControlName = cbxVersion.Name;
        }

        private void edtYear_TextChanged(object sender, EventArgs e) {
            m_lastModControlName = edtYear.Name;
        }

        private void edtFileName_TextChanged(object sender, EventArgs e) {
            m_lastModControlName = edtFileName.Name;
        }

        private void edtTitle_TextChanged(object sender, EventArgs e) {
            m_lastModControlName = edtTitle.Name;
        }

        private void edtAlbum_TextChanged(object sender, EventArgs e) {
            m_lastModControlName = edtAlbum.Name;
        }

        private void edtArtist_TextChanged(object sender, EventArgs e) {
            m_lastModControlName = edtArtist.Name;
        }

        private void edtComment_TextChanged(object sender, EventArgs e) {
            m_lastModControlName = edtComment.Name;
        }

        private void edtTrack_TextChanged(object sender, EventArgs e) {
            m_lastModControlName = edtTrack.Name;
        }

        private void edtTrackMax_TextChanged(object sender, EventArgs e) {
            m_lastModControlName = edtTrackMax.Name;
        }
    }
}