using System;
using System.Collections.Generic;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Globalization;
using System.Collections.Specialized;

namespace ID3TagEditLite {
    public delegate void MainViewRedrawHandler(List<TTagItem> tags);
    public delegate List<TTagItem> GetSelectedTagItemsHandler();
    public delegate void SetUndoHandler(List<TTagItem> tags);
    public delegate void ProgressHandler(int percent);

    public enum UndoMode {
        NON = 0,
        UNDO = 1,
        REDO = 2
    }

    public partial class FormMain : Form {
        public TTagManager m_tagMgr = new TTagManager();
        public TPluginManager m_pluginMgr = new TPluginManager();
        public bool m_bFirstWindow = false;
        Rectangle m_defaultRect;
        Rectangle m_setupRect;
        public string m_originalFormName = "";
        public string m_titleBaseName = "untitled";
        GetSelectedTagItemsHandler m_getSelectedTagItemsHandler;
        MainViewRedrawHandler m_redrawHandler;
        SetUndoHandler m_setUndoHandler;
        List<TTagItem> m_undoTargets = null;
        UndoMode m_undoMode = UndoMode.NON;
        public bool m_bCloseRequested = false;

        public FormMain() {
            InitializeComponent();
            m_originalFormName = this.Text;
            m_tagMgr.m_callbackTargetControl = this;
            m_tagMgr.ProgressCallback = new ProgressHandler(ProgressCallback);
            m_tagMgr.SaveEndCallback = new SaveEndHandler(SaveEndCallback);
            m_tagMgr.MainViewRedrawCallback = new MainViewRedrawHandler(RedrawListViewMain);
            m_defaultRect = this.DesktopBounds;
            if (Screen.PrimaryScreen.WorkingArea.Contains(m_defaultRect) == false)
                m_defaultRect = Screen.PrimaryScreen.WorkingArea;
            m_getSelectedTagItemsHandler = new GetSelectedTagItemsHandler(GetSelectedTagItems);
            m_redrawHandler = new MainViewRedrawHandler(RedrawListViewMain);
            m_setUndoHandler = new SetUndoHandler(SetUndo);
            statusMessage.Text = "";

            listViewMain.Columns[modify.Index].Name = "modify";
            listViewMain.Columns[track.Index].Name = "track";
            listViewMain.Columns[fileName.Index].Name = "file-name";
            listViewMain.Columns[title.Index].Name = "title";
            listViewMain.Columns[artist.Index].Name = "artist";
            listViewMain.Columns[album.Index].Name = "album";
            listViewMain.Columns[year.Index].Name = "year";
            listViewMain.Columns[comment.Index].Name = "comment";
            listViewMain.Columns[genre.Index].Name = "genre";
            listViewMain.Columns[version.Index].Name = "version";
            listViewMain.Columns[loadEncode.Index].Name = "load-encode";
            listViewMain.Columns[saveEncode.Index].Name = "save-encode";

            listViewMain.Columns[modify.Index].Width = Program.thePref.columnWidthModify;
            listViewMain.Columns[track.Index].Width = Program.thePref.columnWidthTrack;
            listViewMain.Columns[fileName.Index].Width = Program.thePref.columnWidthFileName;
            listViewMain.Columns[title.Index].Width = Program.thePref.columnWidthTitle;
            listViewMain.Columns[artist.Index].Width = Program.thePref.columnWidthArtist;
            listViewMain.Columns[album.Index].Width = Program.thePref.columnWidthAlbum;
            listViewMain.Columns[year.Index].Width = Program.thePref.columnWidthYear;
            listViewMain.Columns[comment.Index].Width = Program.thePref.columnWidthComment;
            listViewMain.Columns[genre.Index].Width = Program.thePref.columnWidthGenre;
            listViewMain.Columns[version.Index].Width = Program.thePref.columnWidthVersion;
            listViewMain.Columns[loadEncode.Index].Width = Program.thePref.columnWidthSourceEncode;
            listViewMain.Columns[saveEncode.Index].Width = Program.thePref.columnWidthTargetEncode;
        }

        private void MainForm_Load(object sender, EventArgs e) {
            if (m_setupRect.IsEmpty == false) {
                DesktopBounds = m_setupRect;
            }
            // vOC
            m_pluginMgr.Setup(m_redrawHandler);
            foreach (TPluginItem p in m_pluginMgr.m_plugins.Values) {
                p.m_getSelectedTagItemsCallback = m_getSelectedTagItemsHandler;
                p.m_setUndoCallback = m_setUndoHandler;
                ToolStripMenuItem m = new ToolStripMenuItem(p.GetNameForMenu(), null, new EventHandler(p.OnClick));
                this.pluginToolStripMenuItem.DropDownItems.Add(m);
            }
            // GR[fBO
            saveEncAnsiToolStripMenuItem.Text = "&" + TTagRecord.GetEncodingString(TTagEncoding.ANSI);
            saveEncUnicodeToolStripMenuItem.Text = "&" + TTagRecord.GetEncodingString(TTagEncoding.UTF16);
            saveEncAnsiToolStripMenuItem.Tag = TTagEncoding.ANSI;
            saveEncUnicodeToolStripMenuItem.Tag = TTagEncoding.UTF16;
            // ^Oo[W
            v10ToolStripMenuItem.Tag = TTagVersion.V_1_0;
            v11ToolStripMenuItem.Tag = TTagVersion.V_1_1;
            v22ToolStripMenuItem.Tag = TTagVersion.V_2_2;
            v23ToolStripMenuItem.Tag = TTagVersion.V_2_3;
            v24ToolStripMenuItem.Tag = TTagVersion.V_2_4;

            ChangeTitle(m_titleBaseName);
            FixMenu();
            FixWindowMenu();
        }

        private void MainForm_FormClosed(object sender, FormClosedEventArgs e) {
            if (m_bFirstWindow) {
                Program.thePref.userBounds = this.DesktopBounds;
                Program.thePref.columnWidthModify = listViewMain.Columns[modify.Index].Width;
                Program.thePref.columnWidthTrack = listViewMain.Columns[track.Index].Width;
                Program.thePref.columnWidthFileName = listViewMain.Columns[fileName.Index].Width;
                Program.thePref.columnWidthTitle = listViewMain.Columns[title.Index].Width;
                Program.thePref.columnWidthArtist = listViewMain.Columns[artist.Index].Width;
                Program.thePref.columnWidthAlbum = listViewMain.Columns[album.Index].Width;
                Program.thePref.columnWidthYear = listViewMain.Columns[year.Index].Width;
                Program.thePref.columnWidthComment = listViewMain.Columns[comment.Index].Width;
                Program.thePref.columnWidthGenre = listViewMain.Columns[genre.Index].Width;
                Program.thePref.columnWidthVersion = listViewMain.Columns[version.Index].Width;
                Program.thePref.columnWidthSourceEncode = listViewMain.Columns[loadEncode.Index].Width;
                Program.thePref.columnWidthTargetEncode = listViewMain.Columns[saveEncode.Index].Width;
            }
            Program.DisposeFormMain(this);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e) {
            List<FormMain> list = new List<FormMain>(Program.m_mainFormList);
            foreach (FormMain f in list) {
                f.Close();
            }
        }

        void ChangeTitle(string title) {
            m_titleBaseName = title;
            this.Text = m_originalFormName + " : " + m_titleBaseName;
        }
        
        private void MainForm_DragDrop(object sender, DragEventArgs e) {
            string[] ss = (string[])e.Data.GetData(DataFormats.FileDrop, false);
            OnLoadAnyFiles(ss);
        }

        void OnLoadAnyFiles(string[] files) {
            List<string> mp3s = new List<string>();
            foreach (string path in files) {
                if (File.Exists(path) == false) {
                    if (Directory.Exists(path))
                        mp3s.Add(path);
                }
                else {
                    string ext = Path.GetExtension(path).ToLower();
                    if (ext.Equals(".mp3"))
                        mp3s.Add(path);
                    else if (ext.Equals(".tsv") || ext.Equals(".csv")) {
                        Import(path);
                        return;
                    }
                }
            }
            if (mp3s.Count > 0)
                AddMP3Files(mp3s.ToArray());
            FixMenu();
        }

        private void MainForm_DragEnter(object sender, DragEventArgs e) {
            e.Effect = e.Data.GetDataPresent(DataFormats.FileDrop) ? DragDropEffects.All : DragDropEffects.None;
        }

        string[] SelectAudioFiles() {
            openFileDialog.InitialDirectory = Program.thePref.loadDir;
            openFileDialog.Filter = "MP3 (*.mp3)|*.mp3|All files (*.*)|*.*";
            openFileDialog.DefaultExt = "mp3";
            DialogResult result = openFileDialog.ShowDialog();
            if (result == DialogResult.OK) {
                Program.thePref.loadDir = openFileDialog.InitialDirectory;
                return openFileDialog.FileNames;
            }
            return null;
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e) {
            string[] files = SelectAudioFiles();
            if (files != null && files.Length > 0) {
                FormMain f = Program.NewFormMain(1, DesktopBounds);
                f.AddMP3Files(files);
                FixWindowMenuAll();
            }
            FixMenu();
        }
        
        private void addToolStripMenuItem_Click(object sender, EventArgs e) {
            string[] files = SelectAudioFiles();
            if (files != null && files.Length > 0)
                AddMP3Files(files);
            FixMenu();
        }

        public void AddMP3Files(string[] s) {
            if (s.Length <= 0)
                return;
            bool bSort = m_tagMgr.m_tags.Count == 0;
            List<string> list = new List<string>(s);
            list.Sort();
            s = list.ToArray();
            if (m_tagMgr.m_tags.Count == 0) {
                string path = s[0];
                string name = "";
                if (File.Exists(path) == false && Directory.Exists(path))
                    name = Path.GetFileName(path);
                else
                    name = Path.GetFileName(Path.GetDirectoryName(path));
                ChangeTitle(name);
                FixWindowMenuAll();
            }
            int n = m_tagMgr.AppendFiles(s);
            if (bSort)
                m_tagMgr.m_tags.Sort(new TTagComparer());
            IncreaseListViewMain(n);
            RedrawListViewMain();
        }

        public void RedrawListViewMainSub(TTagItem t) {
            int n = m_tagMgr.m_tags.IndexOf(t);
            if (n < 0)
                return;
            ListViewItem item = listViewMain.Items[n];
            item.SubItems[modify.Index].Text = t.m_record.GetModifyString();
            item.SubItems[track.Index].Text = t.m_record.GetTrackString();
            item.SubItems[fileName.Index].Text = t.m_record.m_fileName;
            item.SubItems[title.Index].Text = t.m_record.m_title.m_str;
            item.SubItems[artist.Index].Text = t.m_record.m_artist.m_str;
            item.SubItems[album.Index].Text = t.m_record.m_album.m_str;
            item.SubItems[year.Index].Text = t.m_record.m_year;
            item.SubItems[comment.Index].Text = t.m_record.m_comment.m_str;
            item.SubItems[genre.Index].Text = t.m_record.GetGenreString();
            item.SubItems[version.Index].Text = t.m_record.GetVersionString();
            item.SubItems[loadEncode.Index].Text = TTagRecord.GetEncodingString(t.m_record.m_loadEnc);
            item.SubItems[saveEncode.Index].Text = TTagRecord.GetEncodingString(t.m_record.m_saveEnc);
        }

        public void RedrawListViewMain(TTagItem t) {
            listViewMain.BeginUpdate();
            RedrawListViewMainSub(t);
            listViewMain.EndUpdate();
            FixMenu();
        }

        public void RedrawListViewMain(List<TTagItem> tags) {
            listViewMain.BeginUpdate();
            foreach (TTagItem t in tags) {
                RedrawListViewMainSub(t);
            }
            listViewMain.EndUpdate();
            FixMenu();
        }

        public void RedrawListViewMain() {
            RedrawListViewMain(m_tagMgr.m_tags);
        }

        void IncreaseListViewMain(int n) {
            for (int i = 0; i < n; i++) {
                ListViewItem item = new ListViewItem("");
                for (int j = 1; j < listViewMain.Columns.Count; j++) {
                    item.SubItems.Add("");
                }
                listViewMain.Items.Add(item);
            }
        }

        private void listViewMain_DoubleClick(object sender, EventArgs e) {
            editTagToolStripMenuItem_Click(sender, e);
        }

        string m_lastModControlName = "";

        void EditWithDlg(List<TTagItem> tags) {
            if (tags == null || tags.Count <= 0)
                return;
            FormEditDlg dlg = new FormEditDlg();
            dlg.m_lastModControlName = m_lastModControlName;
            dlg.m_localLocation = Program.thePref.editDlgLocation;
            dlg.m_parentDesktopLocation = this.DesktopLocation;
            dlg.Setup(tags[0].m_record.Clone(), tags.Count == 1 ? tags[0].m_filePath : "", tags.Count > 1);
            if (dlg.ShowDialog() == DialogResult.OK) {
                TTagRecord r = dlg.GetTagRecord();
                SetUndo(tags);
                if (tags.Count > 1) {
                    foreach (TTagItem t in tags) {
                        t.m_record.SetPropertyOfAlbum(r.GetAlbum());
                        t.m_record.SetPropertyOfGenre(r.GetGenreNumberString());
                        t.m_record.SetPropertyOfArtist(r.GetArtist());
                        t.m_record.SetPropertyOfTrackMax(r.GetTrackMaxString());
                        t.m_record.SetPropertyOfYear(r.GetYear());
                        t.m_record.SetPropertyOfLoadEnc(r.m_loadEnc);
                        t.m_record.SetPropertyOfSaveEnc(r.m_saveEnc);
                        t.m_record.SetPropertyOfVersion(r.m_version);
                    }
                }
                else {
                    tags[0].m_record = r;
                }
                RedrawListViewMain(tags);
            }
            m_lastModControlName = dlg.m_lastModControlName;
            Program.thePref.editDlgLocation = dlg.m_localLocation;
        }
        
        private void editTagToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags == null || tags.Count <= 0)
                return;
            EditWithDlg(tags);
            FixMenu();
        }

        private void editAllToolStripMenuItem_Click(object sender, EventArgs e) {
            if (m_tagMgr.m_tags.Count <= 0)
                return;
            EditWithDlg(m_tagMgr.m_tags);
            FixMenu();
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e) {
            m_tagMgr.SaveFiles();
            FixMenu();
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e) {
            FormMain f = Program.NewFormMain(1, DesktopBounds);
            FixWindowMenuAll();
            FixMenu();
        }

        private void closeToolStripMenuItem_Click(object sender, EventArgs e) {
            this.Close();
        }

        public void SetupNewWindowRect(int n, Rectangle userRect) {
            this.m_setupRect = CalcNewWindowRect(n, userRect);
            this.Visible = true;
        }

        public Rectangle CalcNewWindowRect(int gap, Rectangle userRect) {
            int gapx = 20;
            int gapy = 20;
            Rectangle desktop = Screen.PrimaryScreen.WorkingArea;
            /// ʂOĂ
            if (desktop.Contains(userRect) == false) {
                return m_defaultRect;
            }
            Rectangle r = userRect;
            r.Offset(gapx * gap, gapy * gap);
            // ʂɓĂ
            if (desktop.Contains(r)) {
                return r;
            }
            // ړʉʂnYꍇ
            // ړ\
            int m;
            r = userRect;
            for (m = 0;; m++) {
                r.Offset(gapx, gapy);
                if (desktop.Contains(r) == false)
                    break;
            }
            gap = gap % (m + 1);
            r = userRect;
            r.Offset(gapx * gap, gapy * gap);
            return r;
        }

        private void MainForm_Move(object sender, EventArgs e) {
        }

        private void MainForm_SizeChanged(object sender, EventArgs e) {
            if (m_bFirstWindow)
                Program.thePref.userBounds = this.DesktopBounds;
        }

        string GetExportExtByIndex(int n) {
            if (n == 2)
                return "csv";
            return "tsv";
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e) {
            if (m_tagMgr.IsModified()) {
                DialogResult result = MessageBox.Show(this, "Do You Save? \"" + m_titleBaseName + "\"", "Confirm", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);
                if (result == DialogResult.Cancel) {
                    e.Cancel = true;
                    FixMenu();
                }
                else if (result == DialogResult.Yes) {
                    m_tagMgr.SaveFiles();
                    e.Cancel = true;
                    m_bCloseRequested = true;
                    FixMenu();
                }
            }
        }

        public List<TTagItem> GetSelectedTagItems() {
            IEnumerator en = listViewMain.SelectedItems.GetEnumerator();
            List<TTagItem> list = new List<TTagItem>();
            while (en.MoveNext()) {
                ListViewItem item = (ListViewItem)en.Current;
                int n = listViewMain.Items.IndexOf(item);
                list.Add(m_tagMgr.m_tags[n]);
            }
            if (list.Count == 0 && Program.thePref.nothingIsAll) {
                return m_tagMgr.m_tags;
            }
            return list;
        }

        public void SetUndo(List<TTagItem> tags) {
            m_undoTargets = tags;
            foreach (TTagItem t in m_undoTargets) {
                t.CopyToUndoRecord();
            }
            m_undoMode = UndoMode.UNDO;
        }

        private void undoToolStripMenuItem_Click(object sender, EventArgs e) {
            if (m_undoMode == UndoMode.UNDO) {
                foreach (TTagItem t in m_undoTargets) {
                    t.CopyToRedoRecord();
                    t.CopyFromUndoRecord();
                }
                m_undoMode = UndoMode.REDO;
                undoToolStripMenuItem.Text = "Redo";
            }
            else if (m_undoMode == UndoMode.REDO) {
                foreach (TTagItem t in m_undoTargets) {
                    t.CopyToUndoRecord();
                    t.CopyFromRedoRecord();
                }
                m_undoMode = UndoMode.UNDO;
                undoToolStripMenuItem.Text = "Undo";
            }
            RedrawListViewMain(m_undoTargets);
        }

        private void listViewMain_ColumnClick(object sender, ColumnClickEventArgs e) {
            m_tagMgr.m_sortMethod = listViewMain.Columns[e.Column].Name;
            m_tagMgr.m_tags.Sort(new TTagComparer());
            RedrawListViewMain();
        }

        private void autoTrackNumberToolStripMenuItem_Click(object sender, EventArgs e) {
            int i = 1;
            foreach (TTagItem t in m_tagMgr.m_tags) {
                t.m_record.SetPropertyOfTrack(i.ToString());
                i++;
            }
            RedrawListViewMain();
        }

        private void copyToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags == null || tags.Count == 0)
                return;
            string s = "";
            foreach (TTagItem t in tags) {
                s += t.m_record.GetCSVLine();
                s += "\n";
            }
            try {
                Clipboard.Clear();
                //Clipboard.SetText(s, TextDataFormat.Text);
                //Clipboard.SetText(s, TextDataFormat.CommaSeparatedValue);
                Clipboard.SetText(s, TextDataFormat.UnicodeText);
            }
            catch (Exception ex) {
                ///////////////////////////////////////
                // 
                statusMessage.Text = ex.Message;
            }
            FixMenu();
        }

        private void pasteToolStripMenuItem_Click(object sender, EventArgs e) {
            if (Clipboard.ContainsText()) {
                OnPasteText();
            }
            else if (Clipboard.ContainsFileDropList()) {
                StringCollection sc = Clipboard.GetFileDropList();
                string[] ss = new string[sc.Count];
                for (int i = 0; i < sc.Count; i++) {
                    ss[i] = sc[i];
                }
                OnLoadAnyFiles(ss);
            }
        }

        void OnPasteText() {
            string clip = Clipboard.GetText();
            char[] sep = new char[] { };
            string[] lines = clip.Split(new char[] { '\n', '\r' }, StringSplitOptions.RemoveEmptyEntries);
            if (lines == null || lines.Length == 0)
                return;

            int n = 0;
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags.Count > 0) {
                n = m_tagMgr.m_tags.IndexOf(tags[0]);
            }
            List<TTagItem> target = new List<TTagItem>();
            for (int i = 0; i < lines.Length && n < m_tagMgr.m_tags.Count; i++, n++) {
                target.Add(m_tagMgr.m_tags[n]);
            }
            if (target.Count == 0)
                return;
            SetUndo(target);
            for (int i = 0; i < target.Count; i++) {
                target[i].m_record.SetCSVLine(lines[i]);
            }
            RedrawListViewMain(target);
        }

        private void selectAllToolStripMenuItem_Click(object sender, EventArgs e) {
            for (int i = 0; i < listViewMain.Items.Count; i++) {
                listViewMain.Items[i].Selected = true;
            }
            FixMenu();
        }

        private void statusStripMain_ItemClicked(object sender, ToolStripItemClickedEventArgs e) {
            statusMessage.Text = "";
            FixMenu();
        }

        private void exportToolStripMenuItem_Click(object sender, EventArgs e) {
            if (m_tagMgr.m_tags.Count == 0)
                return;
            exportFileDialog.InitialDirectory = Program.thePref.exportDir;
            exportFileDialog.Filter = "TSV (*.tsv)|*.tsv|CSV (*.csv)|*.csv|All files (*.*)|*.*";
            exportFileDialog.FilterIndex = Program.thePref.exportIndex;
            exportFileDialog.DefaultExt = GetExportExtByIndex(exportFileDialog.FilterIndex);
            exportFileDialog.FileName = m_tagMgr.m_tags[0].m_record.GetAlbum();
            if (exportFileDialog.FileName.Length == 0)
                exportFileDialog.FileName = m_titleBaseName;
            //exportFileDialog.FileName += ("." + exportFileDialog.DefaultExt);
            DialogResult result;
            try {
                result = exportFileDialog.ShowDialog();
            }
            catch {
                exportFileDialog.FileName = m_titleBaseName;
                result = exportFileDialog.ShowDialog();
            }
            if (result == DialogResult.OK) {
                Program.thePref.exportIndex = exportFileDialog.FilterIndex;
                Program.thePref.exportDir = exportFileDialog.InitialDirectory;
                string ext = Path.GetExtension(exportFileDialog.FileName).ToLower();
                m_tagMgr.Export(exportFileDialog.FileName, ext);
            }
            FixMenu();
        }

        private void impotToolStripMenuItem_Click(object sender, EventArgs e) {
            if (m_tagMgr.m_tags.Count == 0)
                return;
            importFileDialog.InitialDirectory = Program.thePref.importDir;
            importFileDialog.Filter = "TSV (*.tsv)|*.tsv|CSV (*.csv)|*.csv|All files (*.*)|*.*";
            importFileDialog.FilterIndex = Program.thePref.importIndex;
            importFileDialog.DefaultExt = GetExportExtByIndex(importFileDialog.FilterIndex);
            //importFileDialog.FileName = m_titleBaseName /*+ "." + saveFileDialog.DefaultExt*/;
            DialogResult result = importFileDialog.ShowDialog();
            if (result == DialogResult.OK) {
                Program.thePref.importIndex = importFileDialog.FilterIndex;
                Program.thePref.importDir = importFileDialog.InitialDirectory;
                Import(importFileDialog.FileName);
            }
            FixMenu();
        }

        void Import(string filePath) {
            string ext = Path.GetExtension(filePath).ToLower();
            SetUndo(m_tagMgr.m_tags);
            m_tagMgr.Import(filePath, ext);
            RedrawListViewMain();
        }

        private void saveEncAnsiToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.m_record.SetPropertyOfSaveEnc(TTagEncoding.ANSI);
            }
            RedrawListViewMain(tags);
        }

        private void saveEncUnicodeToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.m_record.SetPropertyOfSaveEnc(TTagEncoding.UTF16);
                if (t.m_record.m_version < TTagVersion.V_2_2)
                    t.m_record.SetPropertyOfVersion(TTagVersion.V_2_2);
            }
            RedrawListViewMain(tags);
        }

        private void v10ToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.m_record.SetPropertyOfVersion(TTagVersion.V_1_0);
                t.m_record.SetPropertyOfSaveEnc(TTagEncoding.ANSI);
            }
            RedrawListViewMain(tags);
        }

        private void v11ToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.m_record.SetPropertyOfVersion(TTagVersion.V_1_1);
                t.m_record.SetPropertyOfSaveEnc(TTagEncoding.ANSI);
            }
            RedrawListViewMain(tags);
        }

        private void v22ToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.m_record.SetPropertyOfVersion(TTagVersion.V_2_2);
            }
            RedrawListViewMain(tags);
        }

        private void v23ToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.m_record.SetPropertyOfVersion(TTagVersion.V_2_3);
            }
            RedrawListViewMain(tags);
        }

        private void v24ToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.m_record.SetPropertyOfVersion(TTagVersion.V_2_4);
            }
            RedrawListViewMain(tags);
        }

        private void OnSelectGenre(object sender, EventArgs e) {
            ToolStripMenuItem m = (ToolStripMenuItem)sender;
            TGenreItem g = (TGenreItem)m.Tag;
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.m_record.SetPropertyOfGenre(g.m_number.ToString());
            }
            RedrawListViewMain(tags);
        }

        private void OnSelectWindow(object sender, EventArgs e) {
            FormMain w = (FormMain)((ToolStripMenuItem)sender).Tag;
            w.BringToFront();
            FixMenu();
        }

        private void genreToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags.Count == 0)
                return;
            FormGenreDlg dlg = new FormGenreDlg();
            dlg.m_parentDesktopLocation = this.DesktopLocation;
            dlg.m_localLocation = Program.thePref.genreDlgLocation;
            dlg.Setup(tags[0].m_record.m_genre);
            if (dlg.ShowDialog() == DialogResult.OK) {
                SetUndo(tags);
                foreach (TTagItem t in tags) {
                    t.m_record.SetPropertyOfGenre(dlg.GetGenre().ToString());
                }
                RedrawListViewMain(tags);
            }
            Program.thePref.genreDlgLocation = dlg.m_localLocation;
            FixMenu();
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e) {
            FormAboutDlg dlg = new FormAboutDlg();
            dlg.ShowDialog();
            FixMenu();
        }

        public void FixWindowMenu() {
            windowToolStripMenuItem.DropDownItems.Clear();
            foreach (FormMain w in Program.m_mainFormList) {
                ToolStripMenuItem m = new ToolStripMenuItem(w.m_titleBaseName, null, new EventHandler(OnSelectWindow));
                m.Tag = w;
                windowToolStripMenuItem.DropDownItems.Add(m);
            }
        }

        public void FixWindowMenuAll() {
            foreach (FormMain w in Program.m_mainFormList) {
                w.FixWindowMenu();
            }
        }

        private void preferencesToolStripMenuItem_Click(object sender, EventArgs e) {
            FormPreferencesDlg dlg = new FormPreferencesDlg();
            if (dlg.ShowDialog() == DialogResult.OK) {
                foreach (FormMain f in Program.m_mainFormList) {
                    foreach (TTagItem t in f.m_tagMgr.m_tags) {
                        t.m_record.DecodeStrings();
                    }
                    f.RedrawListViewMain();
                }
            }
        }

        public void ProgressCallback(int percent) {
            toolStripProgressBar.Value = percent;
        }

        public void SaveEndCallback() {
            toolStripProgressBar.Value = 0;
            if (m_bCloseRequested) {
                m_bCloseRequested = false;
                Close();
            }
        }

        private void listViewMain_SelectedIndexChanged(object sender, EventArgs e) {
            FixMenu();
        }

        private void menuStripMain_ItemClicked(object sender, ToolStripItemClickedEventArgs e) {
            FixMenu();
        }

        void FixMenu() {
            if (m_undoTargets == null || m_undoTargets.Count == 0)
                undoToolStripMenuItem.Enabled = false;
            else
                undoToolStripMenuItem.Enabled = true;
            if (m_tagMgr.m_tags.Count == 0) {
                copyToolStripMenuItem.Enabled = false;
                iD3ToolStripMenuItem.Enabled = false;
                pluginToolStripMenuItem.Enabled = false;
                saveToolStripMenuItem.Enabled = false;
                exportToolStripMenuItem.Enabled = false;
                impotToolStripMenuItem.Enabled = false;
            }
            else {
                iD3ToolStripMenuItem.Enabled = true;
                editAllToolStripMenuItem.Enabled = true;
                saveToolStripMenuItem.Enabled = true;
                exportToolStripMenuItem.Enabled = true;
                impotToolStripMenuItem.Enabled = true;
                List<TTagItem> tags = GetSelectedTagItems();
                if (tags.Count == 0) {
                    copyToolStripMenuItem.Enabled = false;
                    editTagToolStripMenuItem.Enabled = false;
                    pluginToolStripMenuItem.Enabled = false;
                    saveEncToolStripMenuItem.Enabled = false;
                    versionToolStripMenuItem.Enabled = false;
                    genreToolStripMenuItem.Enabled = false;
                }
                else {
                    copyToolStripMenuItem.Enabled = true;
                    editTagToolStripMenuItem.Enabled = true;
                    pluginToolStripMenuItem.Enabled = true;
                    saveEncToolStripMenuItem.Enabled = true;
                    versionToolStripMenuItem.Enabled = true;
                    genreToolStripMenuItem.Enabled = true;
                    saveEncAnsiToolStripMenuItem.Text = "&" + TTagRecord.GetEncodingString(TTagEncoding.ANSI);
                    saveEncUnicodeToolStripMenuItem.Text = "&" + TTagRecord.GetEncodingString(TTagEncoding.UTF16);
                    if (tags.Count == 1) {
                        TTagItem t = tags[0];
                        foreach (ToolStripMenuItem m in saveEncToolStripMenuItem.DropDownItems) {
                            m.CheckState = t.m_record.m_saveEnc == (TTagEncoding)m.Tag ? CheckState.Checked : CheckState.Unchecked;
                        }
                        foreach (ToolStripMenuItem m in versionToolStripMenuItem.DropDownItems) {
                            m.CheckState = t.m_record.m_version == (TTagVersion)m.Tag ? CheckState.Checked : CheckState.Unchecked;
                        }
                        foreach (ToolStripMenuItem m in genreToolStripMenuItem.DropDownItems) {
                            m.CheckState = t.m_record.m_genre == ((TGenreItem)m.Tag).m_number ? CheckState.Checked : CheckState.Unchecked;
                        }
                    }
                    else {
                        foreach (ToolStripMenuItem m in saveEncToolStripMenuItem.DropDownItems) {
                            m.CheckState = CheckState.Unchecked;
                        }
                        foreach (ToolStripMenuItem m in versionToolStripMenuItem.DropDownItems) {
                            m.CheckState = CheckState.Unchecked;
                        }
                        foreach (ToolStripMenuItem m in genreToolStripMenuItem.DropDownItems) {
                            m.CheckState = CheckState.Unchecked;
                        }
                    }
                }
            }
            if (Clipboard.ContainsText() || Clipboard.ContainsFileDropList()) {
                pasteToolStripMenuItem.Enabled = true;
            }
            else {
                pasteToolStripMenuItem.Enabled = false;
            }
        }

        private void listViewMain_KeyUp(object sender, KeyEventArgs e) {
        }

        private void listViewMain_KeyPress(object sender, KeyPressEventArgs e) {
        }

        private void listViewMain_KeyDown(object sender, KeyEventArgs e) {
            if (e.KeyCode == Keys.Enter) {
                if (e.Shift)
                    editAllToolStripMenuItem_Click(sender, e);
                else
                    editTagToolStripMenuItem_Click(sender, e);
            }

        }
    }
}
