using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Collections;
using System.Diagnostics;

namespace ID3TagEditLite {
    public class TGenreItem {
        public int m_number;
        public string m_name;

        public TGenreItem(int number, string name) {
            m_number = number;
            m_name = name;
        }

        public string GetNameForMenu() {
            return m_number.ToString() + ":" + m_name;
        }
    }

    public class TGenreManager {
        public SortedList<int, TGenreItem> m_genres = new SortedList<int, TGenreItem>();

        public void Setup() {
            string moduleFilePath = Process.GetCurrentProcess().MainModule.FileName;
            string dirPath = Path.GetDirectoryName(moduleFilePath);
            string genreDatFilePath = Path.Combine(dirPath, "genre.info");
            LoadGenreDatFile(genreDatFilePath);
        }

        public bool LoadGenreDatFile(string filePath) {
            if (File.Exists(filePath) == false)
                return false;
            FileStream f = new FileStream(filePath, FileMode.Open, FileAccess.Read);
            if (f == null)
                return false;
            StreamReader sr = new StreamReader(f);
            char[] dlmt = new char[] { '\t' };
            while (true) {
                string s = sr.ReadLine();
                if (s == null)
                    break;
                if (s.Length == 0)
                    continue;
                string[] ss = s.Split(dlmt);
                if (ss.Length < 2)
                    continue;
                int number = 0;
                if (TTools.StringToInteger(ref number, ss[0]) == false || number < 0)
                    continue;
                if (ss[1] == null || ss[1].Length == 0)
                    continue;
                m_genres.Add(number, new TGenreItem(number, ss[1]));
            }

            sr.Close();
            f.Close();
            return true;
        }

        public TGenreItem GetGenreItemByGenreNumber(int n) {
            TGenreItem g = null;
            if (m_genres.ContainsKey(n))
                g = m_genres[n];
            return g;
        }

        public bool GetGenreNameByGenreNumber(ref string name, int genre) {
            TGenreItem g = GetGenreItemByGenreNumber(genre);
            if (g == null)
                return false;
            name = g.m_name;
            return true;
        }

        public bool SearchGenreNumberByGenreName(ref int number, string name) {
            foreach (TGenreItem g in m_genres.Values) {
                if (g.m_name.Equals(name, StringComparison.CurrentCultureIgnoreCase)){
                    number = g.m_number;
                    return true;
                }
            }
            return false;
        }
    }
}
