using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace ID3TagEditLite {

    public class TID3V2Comment {
        public string m_language = "eng";
        public TTagString m_description = new TTagString();
        public TTagString m_text = new TTagString();

        public bool DecodeString(TTagEncoding code) {
            m_description.DecodeString(code);
            return m_text.DecodeString(code);
        }

        public void EncodeString(TTagEncoding code, int limit) {
            m_description.EncodeString(code, limit);
            m_text.EncodeString(code, limit);
        }

        public TID3V2Comment Clone() {
            TID3V2Comment c = new TID3V2Comment();
            c.m_language = (string)m_language.Clone();
            c.m_description = m_description.Clone();
            c.m_text = m_text.Clone();
            return c;
        }
    }

    public class TID3V2Head {
        /*
        struct ID3_V2_HEAD {
            char m_tag[3];
            char m_major;
            char m_minor;
            char m_flags;
            char m_size[4];
        };
        */
        public const int HEAD_SIZE = 10;
        public const int FOOT_SIZE = HEAD_SIZE;
        public const int TAG_V2_HEAD_FLAG_FOOTER = 0x10;
        public const int TAG_V2_HEAD_FLAG_UNSYNC = 0x80;
        public const int TAG_V2_HEAD_FLAG_EXTENDED = 0x40;
        public const int TAG_V2_HEAD_FLAG_EXPERIMENTAL = 0x20;
        //
        public byte[] m_magic;
        public byte m_major = (byte)TTagVersion.V_2_2;
        public byte m_minor = 0;
        public byte m_flags = 0;
        public int m_size = 0;

        public bool Read(BinaryReader br) {
            m_magic = br.ReadBytes(3);
            ASCIIEncoding ae = new ASCIIEncoding();
            if (ae.GetString(m_magic).Equals("ID3") == false)
                return false;
            m_major = br.ReadByte();
            m_minor = br.ReadByte();
            m_flags = br.ReadByte();
            byte[] data = br.ReadBytes(4);
            m_size = TagSizeToInteger(data);
            if (m_major == (byte)TTagVersion.V_2_4 && (m_flags & TAG_V2_HEAD_FLAG_FOOTER) != 0)
                m_size += FOOT_SIZE;
            return true;
        }

        public bool Write(BinaryWriter bw) {
            ASCIIEncoding ae = new ASCIIEncoding();
            m_magic = ae.GetBytes("ID3");
            bw.Write(m_magic);
            bw.Write(m_major);
            bw.Write(m_minor);
            bw.Write(m_flags);
            byte[] sz = TagSizeFromInteger(m_size);
            bw.Write(sz);
            return true;
        }

        int TagSizeToInteger(byte[] data) {
            int n = (data[0] << (7 * 3)) + (data[1] << (7 * 2)) + (data[2] << (7 * 1)) + (data[3]);
            return n;
        }

        byte[] TagSizeFromInteger(int n) {
            byte[] data = new byte[4];
            data[3] = (byte)(n & 0x7F);
            data[2] = (byte)((n >> 7) & 0x7F);
            data[1] = (byte)((n >> 7 * 2) & 0x7F);
            data[0] = (byte)((n >> 7 * 3) & 0x7F);
            return data;
        }
    }

    public class TID3V2Frame {
        public string m_frameID = "";
        public byte[] m_flags = null;
        public int m_bodyBytes = 0;
        public byte[] m_body = new byte[0];
        //
        public TTagVersion m_version = TTagVersion.V_2_2;
        int m_frameIDSize = 3;
        int m_dataSizeSize = 3;
        int m_flagsSize = 0;
        //
        public TID3V2Frame(TTagVersion value) {
            SetVersion(value);
            m_flags = new byte[m_flagsSize];
            for (int i = 0; i < m_flagsSize; i++) {
                m_flags[i] = 0;
            }
        }
        
        public TID3V2Frame Clone() {
            TID3V2Frame f = new TID3V2Frame(m_version);
            f.m_frameID = (string)m_frameID.Clone();
            f.m_bodyBytes = m_bodyBytes;
            f.m_body = (byte[])m_body.Clone();
            f.m_flags = (byte[])m_flags.Clone();
            return f;
        }

        public void SetVersion(TTagVersion value) {
            m_version = value;
            if (m_version == TTagVersion.V_2_2) {
                m_frameIDSize = 3;
                m_dataSizeSize = 3;
                m_flagsSize = 0;
            }
            else {
                m_frameIDSize = 4;
                m_dataSizeSize = 4;
                m_flagsSize = 2;
            }
        }
        
        public TTagString GetAsTagString() {
            TTagString s = new TTagString();
            int count = m_body.Length - 1;
            s.Set(m_body, 1, count, (TTagEncoding)m_body[0]);
            return s;
        }

        public string GetAsAsciiString() {
            if (m_body.Length > 1) {
                int count = m_body.Length - 1;
                if (m_body[1 + count - 1] == 0)
                    count -= 1;
                if (count > 0) {
                    ASCIIEncoding ae = new ASCIIEncoding();
                    return ae.GetString(m_body, 1, count);
                }
            }
            return "";
        }

        public TID3V2Comment GetAsID3V2Comment() {
            ASCIIEncoding ae = new ASCIIEncoding();
            TID3V2Comment c = new TID3V2Comment();
            c.m_language = ae.GetString(m_body, 1, 3);
            // LN^̃TCY
            int term = m_body[0] == (byte)TTagEncoding.ANSI ? 1 : 2;
            // fXNvṼoCg𓾂B
            int count = 0;
            for (int i = 4; ; i += term) {
                if (term == 2) {
                    if (m_body[i] == 0 && m_body[i + 1] == 0)
                        break;
                }
                else {
                    if (m_body[i] == 0)
                        break;
                }
                count += term;
            }

            c.m_description.Set(m_body, 4, count, (TTagEncoding)m_body[0]);
            c.m_text.Set(m_body, 4 + count + term, m_bodyBytes - (4 + count + term), (TTagEncoding)m_body[0]);

            return c;
        }

        public bool Write(BinaryWriter bw) {
            ASCIIEncoding ae = new ASCIIEncoding();
            byte[] frameID = ae.GetBytes(m_frameID);
            bw.Write(frameID);
            byte[] bodyBytes = new byte[m_dataSizeSize];
            for (int i = 0; i < m_dataSizeSize; i++) {
                bodyBytes[i] = (byte)((m_bodyBytes >> (8 * (m_dataSizeSize - 1 - i))) & 0xFF);
            }
            bw.Write(bodyBytes);
            if (m_flagsSize > 0) {
                bw.Write(m_flags);
            }
            bw.Write(m_body);
            return true;
        }

        public bool MakeTextFrame(string frameID, TTagString ts, TTagVersion version) {
            m_frameID = frameID;
            m_bodyBytes = ts.m_org.Length + 1;
            m_body = new byte[m_bodyBytes];
            m_body[0] = (byte)ts.m_encoding;
            Array.Copy(ts.m_org, 0, m_body, 1, ts.m_org.Length);
            m_version = version;
            return true;
        }

        public bool WriteTextFrame(BinaryWriter bw, string frameID, TTagString ts, TTagVersion version) {
            if (MakeTextFrame(frameID, ts, version))
                return Write(bw);
            else
                return false;
        }

        public bool MakeAsciiFrame(string frameID, string s, TTagVersion version) {
            m_frameID = frameID;
            ASCIIEncoding ae = new ASCIIEncoding();
            byte[] bs = ae.GetBytes(s);
            m_bodyBytes = bs.Length + 1;
            m_body = new byte[m_bodyBytes];
            m_body[0] = (byte)TTagEncoding.ANSI;
            Array.Copy(bs, 0, m_body, 1, bs.Length);
            m_version = version;
            return true;
        }

        public bool WriteAsciiFrame(BinaryWriter bw, string frameID, string s, TTagVersion version) {
            if (MakeAsciiFrame(frameID, s, version))
                return Write(bw);
            else
                return false;
        }

        public bool MakeCommentFrame(string frameID, TTagString ts, TTagVersion version) {
            m_frameID = frameID;
            if (ts.m_org.Length == 0)
                return false;
            int descbytes = ts.m_encoding < TTagEncoding.UTF16 ? 1 : 4;
            m_bodyBytes = ts.m_org.Length + 1 /* encoding */ + 3 /* language */ + descbytes;
            m_body = new byte[m_bodyBytes];
            int p = 0;
            m_body[p] = (byte)ts.m_encoding;
            p++;
            ASCIIEncoding ae = new ASCIIEncoding();
            byte[] lang = ae.GetBytes("eng");
            Array.Copy(lang, 0, m_body, p, lang.Length);
            p += lang.Length;
            byte[] desc;
            if (ts.m_encoding < TTagEncoding.UTF16)
                desc = new byte[] { 0 };
            else
                desc = new byte[] { 0xFF, 0xFE, 0, 0 };
            Array.Copy(desc, 0, m_body, p, descbytes);
            p += descbytes;
            Array.Copy(ts.m_org, 0, m_body, p, ts.m_org.Length);
            return true;
        }

        public bool WriteCommentFrame(BinaryWriter bw, string frameID, TTagString ts, TTagVersion version) {
            if (MakeCommentFrame(frameID, ts, version))
                return Write(bw);
            else
                return false;
        }

        public int Read(BinaryReader br) {
            ASCIIEncoding ae = new ASCIIEncoding();
            byte[] data = br.ReadBytes(m_frameIDSize);
            if (data[0] == 0)
                return m_frameIDSize;
            m_frameID = ae.GetString(data);
            data = br.ReadBytes(m_dataSizeSize);
            m_bodyBytes = 0;
            for (int i = 0; i < m_dataSizeSize; i++) {
                m_bodyBytes += (data[i] << (8 * (m_dataSizeSize - 1 - i)));
            }
            //m_bodyBytes = (data[0] << (8 * 2)) + (data[1] << (8 * 1)) + (data[2]);
            if (m_flagsSize > 0) {
                m_flags = br.ReadBytes(m_flagsSize);
            }
            if (m_bodyBytes > 0)
                m_body = br.ReadBytes(m_bodyBytes);
            return m_frameIDSize + m_dataSizeSize + m_flagsSize + m_bodyBytes;
        }
    }

}
