using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;

namespace ID3TagEditLite {
    public class TPluginItem {
        public string m_filePath;
        //public delegate TTagItem[] GetSelectedTagItemsHandler();
        public GetSelectedTagItemsHandler m_getSelectedTagItemsCallback = null;
        public SetUndoHandler m_setUndoCallback = null;
        public TPluginManager m_container;
        
        bool Execute(TTagItem tag) {
            TPluginInfo info = m_container.m_pluginInfos[m_container.GetExtensionForKey(m_filePath)];
            if (info == null)
                return false;
            string exe = info.m_exe;
            if (info.m_exe.Length == 0)
                exe = m_filePath;
            string param = info.m_param;
            if (info.m_exe.Length > 0) {
                if (param.Length > 0)
                    param += " ";
                param += "\"" + m_filePath + "\"";
            }

            List<GetTagStringHandler> getCallbackTable = tag.m_record.m_getTagStringCallbackHandlerTable;
            List<SetTagStringHandler> setCallbackTable = tag.m_record.m_setTagStringCallbackHandlerTable;

            Encoding enc;
            if (m_container.m_asis.Contains(info.m_encoding.ToLower()))
                enc = Encoding.GetEncoding(Program.thePref.ansiCodePage);
            else
                enc = new UTF8Encoding(false, false);

            Process proc = new Process();
            proc.StartInfo.FileName = exe;
            proc.StartInfo.Arguments = param ;
            proc.StartInfo.UseShellExecute = false;
            proc.StartInfo.RedirectStandardInput = true;
            proc.StartInfo.RedirectStandardOutput = true;
            //proc.StartInfo.WindowStyle = ProcessWindowStyle.Minimized;
            proc.StartInfo.CreateNoWindow = true;
            proc.StartInfo.StandardOutputEncoding = enc;
            proc.Start();

            StreamWriter sw = new StreamWriter(proc.StandardInput.BaseStream, enc);
            StreamReader sr = proc.StandardOutput;
            string path = tag.m_filePath;
            path = path.Replace('\\', '/');
            sw.WriteLine(path);
            foreach (GetTagStringHandler getCallback in getCallbackTable) {
                sw.WriteLine(getCallback());
            }
            sw.Close();
            path = sr.ReadLine();
            foreach (SetTagStringHandler setCallback in setCallbackTable) {
                string s = sr.ReadLine();
                if (s == null)
                    break;
                setCallback(s);
            }
            sr.Close();
            proc.WaitForExit();
            proc.Close();
            return true;
        }

        bool IsAsIs(string s) {
            return m_container.m_asis.Contains(s.ToLower());
        }

        public string GetNameForMenu() {
            return Path.GetFileNameWithoutExtension(m_filePath);
        }

        public void OnClick(Object sender, EventArgs e) {
            List<TTagItem> tags = m_getSelectedTagItemsCallback();
            if (tags == null || tags.Count == 0)
                return;
            if (m_setUndoCallback != null)
            m_setUndoCallback(tags);
            foreach (TTagItem t in tags) {
                Execute(t);
            }
            m_container.Redraw(tags);
        }
    }

    public class TPluginInfo {
        public string m_ext = "";
        public string m_exe = "";
        public string m_param = "";
        public string m_encoding = "";
    }

    public class TPluginManager {
        public SortedDictionary<string, TPluginItem> m_plugins = new SortedDictionary<string, TPluginItem>();
        public MainViewRedrawHandler m_redrawCallback;
        public SortedDictionary<string, TPluginInfo> m_pluginInfos = new SortedDictionary<string, TPluginInfo>();
        public List<string> m_asis = new List<string>();

        public void Setup(MainViewRedrawHandler redraw) {
            m_redrawCallback = redraw;
            string moduleFilePath = Process.GetCurrentProcess().MainModule.FileName;
            string dirPath = Path.GetDirectoryName(moduleFilePath);
            string pluginInfoPath = Path.Combine(dirPath, "Plugin.info");
            LoadPluginInfoFile(pluginInfoPath);
            string pluginDirPath = Path.Combine(dirPath, "Plugin");
            ListupPluginDirectory(pluginDirPath);
            m_asis.Add("asis");
            m_asis.Add("as_is");
            m_asis.Add("as is");
            m_asis.Add("as-is");
            m_asis.Add("ansi");
            m_asis.Add("ascii");
            m_asis.Add("jis");
            m_asis.Add("sjis");
            m_asis.Add("shift jis");
            m_asis.Add("shiftjis");
            m_asis.Add("shift_jis");
            m_asis.Add("shift-jis");
            m_asis.Add("windows31j");
            m_asis.Add("nihongo");
            m_asis.Add("kanji");
        }

        public bool LoadPluginInfoFile(string filePath) {
            if (File.Exists(filePath) == false)
                return false;
            FileStream f = new FileStream(filePath, FileMode.Open, FileAccess.Read);
            if (f == null)
                return false;
            StreamReader sr = new StreamReader(f);
            char[] dlmtTab = new char[] { '\t' };
            char[] dlmtComma = new char[] { ',', ' ' };
            while (true) {
                string s = sr.ReadLine();
                if (s == null)
                    break;
                if (s.Length == 0)
                    continue;
                string[] ss = s.Split(dlmtTab);
                if (ss.Length == 0)
                    continue;
                string[] exts = ss[0].Split(dlmtComma, StringSplitOptions.RemoveEmptyEntries);
                List<TPluginInfo> infos = new List<TPluginInfo>();
                foreach (string ext in exts) {
                    TPluginInfo info = new TPluginInfo();
                    info.m_ext = ext;
                    infos.Add(info);
                }
                if (ss.Length > 1) {
                    foreach (TPluginInfo info in infos)
                        info.m_exe = ss[1];
                }
                if (ss.Length > 2) {
                    foreach (TPluginInfo info in infos)
                        info.m_param = ss[2];
                }
                if (ss.Length > 3) {
                    foreach (TPluginInfo info in infos)
                        info.m_encoding = ss[3];
                }
                foreach (TPluginInfo info in infos)
                    m_pluginInfos.Add(info.m_ext, info);
            }
            sr.Close();
            f.Close();
            return true;
        }

        public string GetExtensionForKey(string path) {
            string ext = Path.GetExtension(path);
            if (ext.Length == 0)
                ext = ".";
            return ext;
        }

        bool ListupPluginDirectory(string dirPath) {
            if (Directory.Exists(dirPath) == false)
                return false;
            string[] files = Directory.GetFiles(dirPath);
            foreach (string path in files) {
                string ext = GetExtensionForKey(path);
                if (m_pluginInfos.ContainsKey(ext) == false)
                    continue;
                TPluginItem p = new TPluginItem();
                p.m_filePath = path;
                p.m_container = this;
                string key = p.GetNameForMenu();
                m_plugins.Add(key, p);
            }
            return true;
        }

        TPluginItem GetPluginItemByMenuName(string name) {
            return m_plugins[name];
        }

        public void Redraw(List<TTagItem> tags) {
            if (m_redrawCallback != null)
                m_redrawCallback(tags);
        }

    }
}
