using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Collections;
using System.Globalization;
using System.Windows.Forms;
using Uguu.IO.Csv;
using System.Threading;

namespace ID3TagEditLite {

    public delegate string GetTagStringHandler();
    public delegate void SetTagStringHandler(string s);
    public delegate void SaveEndHandler();

    /// <summary>
    /// GR[fBO̎ށB
    /// </summary>
    public enum TTagEncoding {
//        LATIN1 = -1,    // tO
        ANSI = 0,      // ID3 Tag V2
        UTF16 = 1
    }

    public enum TTagVersion {
        V_1_0 = 0,      // tO
        V_1_1 = 1,
        V_2_2 = 2,      // ID2 Tag V2
        V_2_3 = 3,
        V_2_4 = 4
    }

    /// <summary>
    /// Undo/Redo₷邽߂̃^Of[^̃NXB
    /// </summary>
    public class TTagRecord {
        public bool m_modify = false;
        public string m_fileName = "";
        public TTagString m_title = new TTagString();
        public TTagString m_artist = new TTagString();
        public TTagString m_album = new TTagString();
        public TTagString m_comment = new TTagString();
        public string m_year = "";
        public int m_genre = 0;
        public TTagVersion m_version = TTagVersion.V_2_2;
        public int m_track = 0;
        public int m_trackMax = 0;
        public TTagEncoding m_loadEnc = (TTagEncoding)Program.thePref.defaultSaveEnc;
        public TTagEncoding m_saveEnc = (TTagEncoding)Program.thePref.defaultSaveEnc;
        // fXNvVRgB
        public List<TID3V2Frame> m_systemComment = new List<TID3V2Frame>();
        // ̑
        public List<TID3V2Frame> m_other = new List<TID3V2Frame>();

        public List<GetTagStringHandler> m_getTagStringCallbackHandlerTable;
        public List<SetTagStringHandler> m_setTagStringCallbackHandlerTable;

        public TTagRecord() {
            m_getTagStringCallbackHandlerTable = CreateGetStringCallbackTable();
            m_setTagStringCallbackHandlerTable = CreateSetStringCallbackTable();
        }

        public TTagRecord Clone() {
            TTagRecord r = new TTagRecord();
            r.m_fileName = (string)m_fileName.Clone();
            r.m_title = m_title.Clone();
            r.m_artist = m_artist.Clone();
            r.m_album = m_album.Clone();
            r.m_comment = m_comment.Clone();
            r.m_year = (string)m_year.Clone();
            r.m_genre = m_genre;
            r.m_version = m_version;
            r.m_track = m_track;
            r.m_trackMax = m_trackMax;
            r.m_loadEnc = m_loadEnc;
            r.m_saveEnc = m_saveEnc;
            foreach (TID3V2Frame c in m_systemComment)
                r.m_systemComment.Add(c.Clone());
            foreach (TID3V2Frame f in m_other)
                r.m_other.Add(f.Clone());
            r.m_modify = m_modify;
            return r;
        }

        public bool DecodeStrings() {
            Boolean mod = false;
            if (m_title.DecodeString(m_loadEnc))
                mod = true;
            if (m_artist.DecodeString(m_loadEnc))
                mod = true;
            if (m_album.DecodeString(m_loadEnc))
                mod = true;
            if (m_comment.DecodeString(m_loadEnc))
                mod = true;
            return mod;
        }

        public bool EncodeStrings() {
            if (m_version >= TTagVersion.V_2_2) {
                m_title.EncodeString(m_saveEnc, -1);
                m_artist.EncodeString(m_saveEnc, -1);
                m_album.EncodeString(m_saveEnc, -1);
                m_comment.EncodeString(m_saveEnc, -1);
            }
            else {
                m_title.EncodeString(m_saveEnc, 30);
                m_artist.EncodeString(m_saveEnc, 30);
                m_album.EncodeString(m_saveEnc, 30);
                m_comment.EncodeString(m_saveEnc, m_version == TTagVersion.V_1_1 ? 28 : 30);
            }
            return true;
        }

        public string GetModifyString() {
            return m_modify ? "+" : "-";
        }

        public string GetFileName() {
            return m_fileName;
        }

        public string GetTitle() {
            return m_title.m_str;
        }

        public string GetArtist() {
            return m_artist.m_str;
        }
        
        public string GetAlbum() {
            return m_album.m_str;
        }

        public string GetYear() {
            return m_year;
        }

        public string GetComment() {
            return m_comment.m_str;
        }

        public string GetVersionString() {
            if ((int)m_version >= 0 && (int)m_version < 5)
                return GetVersionNames()[(int)m_version];
            else
                return "-";
        }

        public string GetGenreNumberString() {
            return m_genre.ToString();
        }

        public string GetGenreString() {
            string g = "";
            if (Program.theGenre.GetGenreNameByGenreNumber(ref g, m_genre) && g.Length > 0)
                return g;
            else
                return GetGenreNumberString();
        }

        public string GetGenreExportString() {
            return "(" + m_genre.ToString() + ")";
        }

        public string GetTrackAndTrackMaxString() {
            if (m_trackMax > 0)
                return GetTrackString() + "/" + GetTrackMaxString();
            else
                return GetTrackString();
        }

        public string GetTrackString() {
            if (m_track > 0)
                return m_track.ToString();
            else
                return "";
        }

        public string GetTrackMaxString() {
            if (m_trackMax > 0)
                return m_trackMax.ToString();
            else
                return "";
        }

        public static string GetEncodingString(TTagEncoding coding) {
//            if (coding == TTagEncoding.LATIN1) {
//                return "LATIN-1";
//            }
            Encoding enc;
            if (coding == TTagEncoding.UTF16) {
                enc = Encoding.Unicode;
            }
            else {
                enc = Encoding.GetEncoding(Program.thePref.ansiCodePage);
            }
            return enc.WebName.ToUpper();
        }

        public static string[] GetEncodingNames() {
            string[] ss = new string[2];
            ss[0] = GetEncodingString(TTagEncoding.ANSI);
            ss[1] = GetEncodingString(TTagEncoding.UTF16);
            return ss;
        }

        public static string[] GetVersionNames() {
            return new string[] { "v1.0", "v1.1", "v2.2", "v2.3", "v2.4" };
        }

        public static string CorrectYear(string year) {
            int y = 0;
            if (TTools.StringToInteger(ref y, year) && y > 0)
                return y.ToString();
            return "";
        }

        public void SetPropertyOfGenre(string value) {
            char[] dlmt = { '(', ')' };
            string[] ss = value.Split(dlmt, StringSplitOptions.RemoveEmptyEntries);
            if (ss.Length > 0)
                SetIntegerProperty(ref m_genre, ss[0]);
        }

        public void SetPropertyOfTrack(string value) {
            SetIntegerProperty(ref m_track, value);
        }

        public void SetPropertyOfTrackMax(string value) {
            SetIntegerProperty(ref m_trackMax, value);
        }

        public void SetPropertyOfYear(string value) {
            string ys = CorrectYear(value);
            SetStringProperty(ref m_year, ys);
        }

        public void SetPropertyOfFileName(string value) {
            SetStringProperty(ref m_fileName, value);
        }

        public void SetPropertyOfTitle(string value) {
            SetTagStringProperty(ref m_title, value);
        }

        public void SetPropertyOfArtist(string value) {
            SetTagStringProperty(ref m_artist, value);
        }

        public void SetPropertyOfAlbum(string value) {
            SetTagStringProperty(ref m_album, value);
        }

        public void SetPropertyOfComment(string value) {
            SetTagStringProperty(ref m_comment, value);
        }

        public void SetPropertyOfVersion(TTagVersion value) {
            if (m_version != value) {
                m_version = value;
                m_modify = true;
            }
        }

        public void SetPropertyOfSaveEnc(TTagEncoding value) {
            if (m_saveEnc != value) {
                m_saveEnc = value;
                m_modify = true;
            }
        }

        public void SetPropertyOfLoadEnc(TTagEncoding value) {
            if (m_loadEnc != value) {
                m_loadEnc = value;
                m_modify = true;
            }
        }

        public void SetStringProperty(ref string var, string value) {
            if (var.Equals(value) == false) {
                var = value;
                m_modify = true;
            }
        }

        public void SetTagStringProperty(ref TTagString var, string value) {
            if (var.m_str.Equals(value) == false) {
                var.m_str = value;
                m_modify = true;
            }
        }

        public void SetIntegerProperty(ref int var, string value) {
            int n = 0;
            if (value.Length == 0) {
                var = n;
                m_modify = true;
            }
            else if (TTools.StringToInteger(ref n, value)) {
                if (n != var) {
                    var = n;
                    m_modify = true;
                }
            }
        }

        List<GetTagStringHandler> CreateGetStringCallbackTable() {
            List<GetTagStringHandler> table = new List<GetTagStringHandler>();
            table.Add(new GetTagStringHandler(GetTrackString));
            table.Add(new GetTagStringHandler(GetFileName));
            table.Add(new GetTagStringHandler(GetTitle));
            table.Add(new GetTagStringHandler(GetAlbum));
            table.Add(new GetTagStringHandler(GetArtist));
            table.Add(new GetTagStringHandler(GetYear));
            table.Add(new GetTagStringHandler(GetComment));
            table.Add(new GetTagStringHandler(GetGenreExportString));
            return table;
        }

        List<SetTagStringHandler> CreateSetStringCallbackTable() {
            List<SetTagStringHandler> table = new List<SetTagStringHandler>();
            table.Add(new SetTagStringHandler(SetPropertyOfTrack));
            table.Add(new SetTagStringHandler(SetPropertyOfFileName));
            table.Add(new SetTagStringHandler(SetPropertyOfTitle));
            table.Add(new SetTagStringHandler(SetPropertyOfAlbum));
            table.Add(new SetTagStringHandler(SetPropertyOfArtist));
            table.Add(new SetTagStringHandler(SetPropertyOfYear));
            table.Add(new SetTagStringHandler(SetPropertyOfComment));
            table.Add(new SetTagStringHandler(SetPropertyOfGenre));
            return table;
        }

        public string GetTSVLine() {
            List<GetTagStringHandler> callbackTable = m_getTagStringCallbackHandlerTable;
            bool bFirst = true;
            string result = "";
            foreach (GetTagStringHandler callback in callbackTable) {
                string s = callback();
                if (bFirst == false)
                    result += "\t";
                result += s;
                bFirst = false;
            }
            return result;
        }

        public void SetTSVLine(string s) {
            List<SetTagStringHandler> callbackTable = m_setTagStringCallbackHandlerTable;
            char[] dlmt = new char[] { '\t' };
            string[] data = s.Split(dlmt);
            for (int i = 0; i < callbackTable.Count && i < data.Length; i++) {
                callbackTable[i](data[i]);
            }
        }

        public string GetCSVLine() {
            UnicodeEncoding enc = new UnicodeEncoding(false, false);
            MemoryStream ms = new MemoryStream();
            StreamWriter sw = new StreamWriter(ms, enc);
            CsvWriter cw = new CsvWriter(sw);
            List<GetTagStringHandler> getCallbackTable = m_getTagStringCallbackHandlerTable;
            foreach (GetTagStringHandler getCallback in getCallbackTable) {
                cw.Write(getCallback());
            }
            cw.Close();
            sw.Close();
            ms.Close();
            byte[] data = ms.ToArray();
            string s = enc.GetString(data);
            return s;
        }

        public void SetCSVLine(string s) {
            UnicodeEncoding enc = new UnicodeEncoding(false, false);
            MemoryStream ms = new MemoryStream(enc.GetBytes(s));
            StreamReader sr = new StreamReader(ms, enc);
            CsvReader cr = new CsvReader(sr);
            List<SetTagStringHandler> setCallbackTable = m_setTagStringCallbackHandlerTable;
            if (cr.Read() == false)
                return;
            for (int i = 0; i < cr.FieldCount && i < setCallbackTable.Count; i++) {
                setCallbackTable[i](cr[i]);
            }
            cr.Close();
            sr.Close();
            ms.Close();
        }
    }

    /// <summary>
    /// 1MP3t@CɊ܂܂^OŜ̃NXB
    /// </summary>
    public class TTagItem {
        public TTagRecord m_record = new TTagRecord();
        public string m_filePath = "";
        public long m_soundStartPos = 0;
        public long m_soundSize = 0;
        public TTagRecord m_undoRecord = null;
        public TTagRecord m_redoRecord = null;
        public string m_tempFilePath = "";
        public TTagManager m_container = null;

        public TTagItem(TTagManager container) {
            m_container = container;
        }
        
        /// <summary>
        /// t@CpXw肵ID3^OǂݍށB
        /// </summary>
        /// <param name="filePath"></param>
        /// <returns></returns>
        public bool LoadTag(string filePath) {
            m_filePath = filePath;
            m_record.m_fileName = Path.GetFileName(m_filePath);
            m_soundStartPos = 0;
            try {
                FileInfo info = new FileInfo(m_filePath);
                m_soundSize = info.Length;
            }
            catch {
                return false;
            }

            LoadV1Tag();
            LoadV2Tag();

            m_record.m_modify = false;

            return true;
        }

        bool LoadV1Tag() {
            FileInfo info = new FileInfo(m_filePath);
            if (info.Length < TID3V1Tag.TAG_SIZE)
                return false;
            FileStream fs = null;
            BinaryReader br = null;
            try {
                fs = new FileStream(m_filePath, FileMode.Open, FileAccess.Read);
                br = new BinaryReader(fs);
            }
            catch {
                return false;
            }
            TID3V1Tag tag = new TID3V1Tag();
            br.BaseStream.Seek(info.Length - TID3V1Tag.TAG_SIZE, SeekOrigin.Begin);
            bool result = tag.Read(br);
            br.Close();
            fs.Close();
            if (result == false)
                return false;

            m_record.m_version = tag.m_null == 0 ? TTagVersion.V_1_1 : TTagVersion.V_1_0;

            ASCIIEncoding ae = new ASCIIEncoding();
            m_record.m_title.Set(tag.m_title, 0, tag.m_title.Length, TTagEncoding.ANSI);
            m_record.m_artist.Set(tag.m_artist, 0, tag.m_artist.Length, TTagEncoding.ANSI);
            m_record.m_album.Set(tag.m_album, 0, tag.m_album.Length, TTagEncoding.ANSI);
            m_record.m_year = TTagRecord.CorrectYear(ae.GetString(tag.m_year));
            m_record.m_comment.Set(tag.m_comment, 0, tag.m_comment.Length, TTagEncoding.ANSI);
            m_record.m_track = tag.m_track;
            m_record.m_genre = tag.m_genre;

            m_soundSize = info.Length - TID3V1Tag.TAG_SIZE;
            m_record.m_loadEnc = TTagEncoding.ANSI;
            m_record.m_saveEnc = m_record.m_loadEnc;
            m_record.DecodeStrings();

            return true;
        }
        
        bool LoadV2Tag() {
            FileInfo info = new FileInfo(m_filePath);
            if (info.Length < TID3V2Head.HEAD_SIZE)
                return false;
            FileStream fs = null;
            BinaryReader br = null;
            try {
                fs = new FileStream(m_filePath, FileMode.Open, FileAccess.Read);
                br = new BinaryReader(fs);
            }
            catch {
                return false;
            }
            TID3V2Head head = new TID3V2Head();
            bool result = head.Read(br);
            if (result) {
                m_soundStartPos = TID3V2Head.HEAD_SIZE + head.m_size;
                m_soundSize -= m_soundStartPos;
                m_record.m_version = (TTagVersion)head.m_major;
                result = LoadV2TagSub(br, head.m_size);
            }
            br.Close();
            fs.Close();
            if (result == false)
                return false;

            // łUNICODEƐݒ肵ĂȂ΁AŜUNICODEňB
            // Ŝߔ > 0 ƂĂB
            if (m_record.m_title.m_encoding > TTagEncoding.ANSI ||
                    m_record.m_artist.m_encoding > TTagEncoding.ANSI ||
                    m_record.m_album.m_encoding > TTagEncoding.ANSI ||
                    m_record.m_comment.m_encoding > TTagEncoding.ANSI) {
                m_record.m_loadEnc = TTagEncoding.UTF16;
            }
            else {
                /* Ȃ`FbNKv */
                m_record.m_loadEnc = TTagEncoding.ANSI;
                if (m_record.m_loadEnc == TTagEncoding.ANSI)
                    m_record.m_loadEnc = m_record.m_title.CheckString();
                if (m_record.m_loadEnc == TTagEncoding.ANSI)
                    m_record.m_loadEnc = m_record.m_artist.CheckString();
                if (m_record.m_loadEnc == TTagEncoding.ANSI)
                    m_record.m_loadEnc = m_record.m_album.CheckString();
                if (m_record.m_loadEnc == TTagEncoding.ANSI)
                    m_record.m_loadEnc = m_record.m_comment.CheckString();
            }

            m_record.m_saveEnc = m_record.m_loadEnc;
            m_record.DecodeStrings();

            return true;
        }

        bool LoadV2TagSub(BinaryReader br, int tagSize) {
            int pos = TID3V2Head.HEAD_SIZE;
            while (pos < tagSize) {
                TID3V2Frame frame = new TID3V2Frame(m_record.m_version);
                pos += frame.Read(br);
                if (frame.m_frameID.Length == 0) {
                    break;
                }
                else if (frame.m_frameID.Equals("TT2") || frame.m_frameID.Equals("TIT2")) {
                    m_record.m_title = frame.GetAsTagString();
                }
                else if (frame.m_frameID.Equals("TP1") || frame.m_frameID.Equals("TPE1")) {
                    m_record.m_artist = frame.GetAsTagString();
                }
                else if (frame.m_frameID.Equals("TAL") || frame.m_frameID.Equals("TALB")) {
                    m_record.m_album = frame.GetAsTagString();
                }
                else if (frame.m_frameID.Equals("TYE") || frame.m_frameID.Equals("TYER")) {
                    m_record.m_year = frame.GetAsAsciiString();
                }
                else if (frame.m_frameID.Equals("COM") || frame.m_frameID.Equals("COMM")) {
                    TID3V2Comment c = frame.GetAsID3V2Comment();
                    c.DecodeString(c.m_text.m_encoding);
                    if (c.m_description.m_str.Length > 0)
                        m_record.m_systemComment.Add(frame);
                    else
                        m_record.m_comment = c.m_text;
                }
                else if (frame.m_frameID.Equals("TCO") || frame.m_frameID.Equals("TCON")) {
                    string[] t = frame.GetAsAsciiString().Split(new char[] { '(', ')' }, StringSplitOptions.RemoveEmptyEntries);
                    if (t.Length > 0) {
                        if (TTools.StringToInteger(ref m_record.m_genre, t[0]) == false) {
                            string[] u = t[0].Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
                            foreach (string s in u) {
                                if (Program.theGenre.SearchGenreNumberByGenreName(ref m_record.m_genre, s))
                                    break;
                            }
                        }
                    }
                }
                else if (frame.m_frameID.Equals("TRK") || frame.m_frameID.Equals("TRCK")) {
                    string[] t = frame.GetAsAsciiString().Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
                    if (t.Length > 0)
                        TTools.StringToInteger(ref m_record.m_track, t[0]);
                    if (t.Length > 1)
                        TTools.StringToInteger(ref m_record.m_trackMax, t[1]);
                }
                else {
                    m_record.m_other.Add(frame);
                }
            }

            return true;
        }

        public byte[] MakeTag() {
            m_record.EncodeStrings();
            m_record.m_year = TTagRecord.CorrectYear(m_record.m_year);
            if (m_record.m_version <= TTagVersion.V_1_1) {
		        return MakeV1Tag();
	        }
	        else {
		        return MakeV2Tag();
	        }
        }

        byte[] FillSpace(byte[] inData, int max) {
            byte[] result = new byte[max];
            for (int i = 0; i < inData.Length; i++) {
                result[i] = inData[i];
            }
            for (int i = inData.Length; i < max; i++) {
                result[i] = 0x20;
            }
            return result;
        }

        byte[] MakeV1Tag() {
            MemoryStream ms = new MemoryStream();
            BinaryWriter bw = new BinaryWriter(ms);
            ASCIIEncoding ae = new ASCIIEncoding();
            bw.Write(ae.GetBytes("TAG"));
            bw.Write(FillSpace(m_record.m_title.m_org, 30));
            bw.Write(FillSpace(m_record.m_artist.m_org, 30));
            bw.Write(FillSpace(m_record.m_album.m_org, 30));
            bw.Write(FillSpace(ae.GetBytes(m_record.m_year), 4));
            if (m_record.m_version == TTagVersion.V_1_1) {
                bw.Write(FillSpace(m_record.m_comment.m_org, 28));
                bw.Write((byte)0);
                bw.Write((byte)m_record.m_track);
            }
            else {
                bw.Write(FillSpace(m_record.m_comment.m_org, 30));
            }
            bw.Write((byte)m_record.m_genre);

            byte[] result = ms.ToArray();
            bw.Close();
            ms.Close();
            return result;
        }

        byte[] MakeV2Tag() {
            MemoryStream ms = new MemoryStream();
            BinaryWriter bw = new BinaryWriter(ms);

            TID3V2Head head = new TID3V2Head();
            head.m_major = (byte)m_record.m_version;
            head.Write(bw);

        	Boolean isV22 = (m_record.m_version == TTagVersion.V_2_2);
            TID3V2Frame frame = new TID3V2Frame(m_record.m_version);
            frame.WriteTextFrame(bw, isV22 ? "TT2" : "TIT2", m_record.m_title, m_record.m_version);
            frame.WriteTextFrame(bw, isV22 ? "TP1" : "TPE1", m_record.m_artist, m_record.m_version);
            frame.WriteTextFrame(bw, isV22 ? "TAL" : "TALB", m_record.m_album, m_record.m_version);
            frame.WriteAsciiFrame(bw, isV22 ? "TCO" : "TCON", m_record.GetGenreExportString(), m_record.m_version);
            frame.WriteAsciiFrame(bw, isV22 ? "TRK" : "TRCK", m_record.GetTrackAndTrackMaxString(), m_record.m_version);
            if (m_record.m_year.Length > 0)
                frame.WriteAsciiFrame(bw, isV22 ? "TYE" : "TYER", m_record.m_year, m_record.m_version);

            foreach (TID3V2Frame f in m_record.m_other) {
                if (f.m_version == m_record.m_version)
                    f.Write(bw);
            }
            foreach (TID3V2Frame c in m_record.m_systemComment) {
                if (c.m_version == m_record.m_version)
                    c.Write(bw);
            }
            frame.WriteCommentFrame(bw, isV22 ? "COM" : "COMM", m_record.m_comment, m_record.m_version);

            long pos = bw.Seek(0, SeekOrigin.Current);
            long sz = 0;
            while (sz < pos + 64) {
                sz += 64;
            }
            long fill = sz - pos;
            for (long i = 0; i < fill; i++)
                bw.Write((byte)0);

            bw.Seek(0, SeekOrigin.Begin);
            head.m_size = (int)sz;
            head.Write(bw);

            byte[] result = ms.ToArray();
            bw.Close();
            ms.Close();
            return result;
        }

        public void CopyToUndoRecord() {
            m_undoRecord = m_record.Clone();
        }

        public void CopyToRedoRecord() {
            m_redoRecord = m_record.Clone();
        }

        public void CopyFromUndoRecord() {
            m_record = m_undoRecord;
        }

        public void CopyFromRedoRecord() {
            m_record = m_redoRecord;
        }

    }

    public class TTagComparer : IComparer<TTagItem> {
        public int Compare(TTagItem x, TTagItem y) {
            int result = 0;
            if (x.m_container.m_sortMethod.Equals("file-name")) {
                result = x.m_record.GetFileName().CompareTo(y.m_record.GetFileName());
            }
            else if (x.m_container.m_sortMethod.Equals("title")) {
                result = x.m_record.GetTitle().CompareTo(y.m_record.GetTitle());
            }
            else if (x.m_container.m_sortMethod.Equals("album")) {
                result = x.m_record.GetAlbum().CompareTo(y.m_record.GetAlbum());
            }
            else if (x.m_container.m_sortMethod.Equals("artist")) {
                result = x.m_record.GetArtist().CompareTo(y.m_record.GetArtist());
            }
            else if (x.m_container.m_sortMethod.Equals("comment")) {
                result = x.m_record.GetComment().CompareTo(y.m_record.GetComment());
            }
            else if (x.m_container.m_sortMethod.Equals("genre")) {
                if (x.m_record.m_genre < y.m_record.m_genre)
                    result = -1;
                else if (x.m_record.m_genre > y.m_record.m_genre)
                    result = 1;
            }
            else if (x.m_container.m_sortMethod.Equals("year")) {
                result = x.m_record.GetYear().CompareTo(y.m_record.GetYear());
            }
            else if (x.m_container.m_sortMethod.Equals("track")) {
                if (x.m_record.m_track < y.m_record.m_track)
                    result = -1;
                else if (x.m_record.m_track > y.m_record.m_track)
                    result = 1;
            }
            else if (x.m_container.m_sortMethod.Equals("version")) {
                if (x.m_record.m_version < y.m_record.m_version)
                    result = -1;
                else if (x.m_record.m_version > y.m_record.m_version)
                    result = 1;
            }
            else if (x.m_container.m_sortMethod.Equals("load-encode")) {
                if (x.m_record.m_loadEnc < y.m_record.m_loadEnc)
                    result = -1;
                else if (x.m_record.m_loadEnc > y.m_record.m_loadEnc)
                    result = 1;
            }
            else if (x.m_container.m_sortMethod.Equals("save-encode")) {
                if (x.m_record.m_saveEnc < y.m_record.m_saveEnc)
                    result = -1;
                else if (x.m_record.m_saveEnc > y.m_record.m_saveEnc)
                    result = 1;
            }

            if (result != 0)
                return result;

            result = x.m_record.GetFileName().CompareTo(y.m_record.GetFileName());
            if (result == 0)
                result = x.m_record.GetTitle().CompareTo(y.m_record.GetTitle());
            if (result == 0) {
                if (x.m_record.m_track < y.m_record.m_track)
                    result = -1;
                else if (x.m_record.m_track > y.m_record.m_track)
                    result = 1;
            }
            if (result == 0)
                result = x.m_record.GetAlbum().CompareTo(y.m_record.GetAlbum());
            if (result == 0)
                result = x.m_record.GetArtist().CompareTo(y.m_record.GetArtist());

            return result;
        }
    }

    public class TTagManager {
        //public delegate void UpdateLVMainCallbackHandler(int n);
        public MainViewRedrawHandler MainViewRedrawCallback = null;
        public List<TTagItem> m_tags = new List<TTagItem>();
        public string m_sortMethod = "track";

		public Control m_callbackTargetControl = null;
		public ProgressHandler ProgressCallback = null;
        public SaveEndHandler SaveEndCallback = null;

        Thread m_work = null;
        bool m_continue = false;

        TTagItem GetByFilePath(string path) {
            foreach (TTagItem t in m_tags) {
                if (t.m_filePath.Equals(path, StringComparison.CurrentCultureIgnoreCase))
                    return t;
            }
            return null;
        }
        
        public int AppendFiles(string[] pathArray) {
            int result = 0;
            if (pathArray == null || pathArray.Length == 0)
                return result;
            foreach (string path in pathArray) {
                result += LoadTagFiles(path);
            }
            return result;
        }

        public int LoadTagFiles(string path) {
            int result = 0;
            if (Directory.Exists(path)) {
                DirectoryInfo info = new DirectoryInfo(path);
                FileInfo[] files = info.GetFiles();
                foreach (FileInfo file in files) {
                    string ext = Path.GetExtension(file.FullName).ToLower();
                    if (ext.Equals(".mp3")) {
                        if (GetByFilePath(file.FullName) == null) {
                            TTagItem t = new TTagItem(this);
                            if (t.LoadTag(file.FullName)) {
                                m_tags.Add(t);
                                result++;
                            }
                        }
                    }
                }
                Program.thePref.loadDir = path;
            }
            else if (File.Exists(path)) {
                FileInfo info = new FileInfo(path);
                string ext = Path.GetExtension(path).ToLower();
                if (ext.Equals(".mp3")) {
                    if (GetByFilePath(path) == null) {
                        TTagItem t = new TTagItem(this);
                        if (t.LoadTag(path)) {
                            m_tags.Add(t);
                            result++;
                        }
                        Program.thePref.loadDir = info.DirectoryName;
                    }
                }
            }
            return result;
        }

        public bool IsWorking() {
            return m_work != null;
        }

        public void Stop() {
            lock (this) {
                m_continue = false;
            }
        }

        public void SaveFiles() {
            int count = m_tags.Count;
            if (count == 0)
                return;
            m_continue = true;
            m_work = new Thread(new ThreadStart(SaveFilesThread));
            m_work.Start();
        }

        public void SaveFilesThread() {
            int count = m_tags.Count;
            bool[] bSaveTemp = new bool[count];

            long soundSizeAll = 0;
            for (int index = 0; index < count; index++) {
                TTagItem tag = m_tags[index];
                soundSizeAll += tag.m_soundSize;
            }
            long soundSizeDone = 0;
            int percent = 0;

            for (int index = 0; index < count; index++) {
                bSaveTemp[index] = false;
                TTagItem tag = m_tags[index];
                if (tag.m_record.m_modify == false)
                    continue;
                byte[] tagData = tag.MakeTag();
                if (tagData.Length == 0)
                    continue;
                if (tag.m_record.m_version >= TTagVersion.V_2_2)
                    bSaveTemp[index] = true;
                if (tag.m_record.m_fileName.Equals(Path.GetFileName(tag.m_filePath)) == false)
                    bSaveTemp[index] = true;
                if (tag.m_soundStartPos != 0)
                    bSaveTemp[index] = true;

                if (bSaveTemp[index] == false) {
                    // V1.x only
                    FileStream fsOut = new FileStream(tag.m_filePath, FileMode.Open, FileAccess.ReadWrite);
                    BinaryWriter bw = new BinaryWriter(fsOut);
                    bw.BaseStream.SetLength(tag.m_soundSize);
                    bw.BaseStream.Seek(0, SeekOrigin.End);
                    bw.Write(tagData);
                    bw.Close();
                    fsOut.Close();
                    soundSizeDone += tag.m_soundSize;
                    percent = (int)(100 * soundSizeDone / soundSizeAll);
                    ProgressCallbackThread(percent);
                }
                else {
                    // V1.x or V2.x
                    tag.m_tempFilePath = tag.m_filePath + ".~temp~";
                    FileStream fsIn = new FileStream(tag.m_filePath, FileMode.Open, FileAccess.Read);
                    BufferedStream br = new BufferedStream(fsIn, 64 * 1024);
                    FileStream fsOut = new FileStream(tag.m_tempFilePath, FileMode.Create, FileAccess.Write);
                    BufferedStream bsOut = new BufferedStream(fsOut, 64 * 1024);
                    BinaryWriter bw = new BinaryWriter(bsOut);

                    if (tag.m_record.m_version >= TTagVersion.V_2_2)
                        bw.Write(tagData);
                    br.Seek(tag.m_soundStartPos, SeekOrigin.Begin);
                    for (int i = 0; i < tag.m_soundSize; i++) {
                        bw.Write((byte)br.ReadByte());
                        soundSizeDone++;
                        int p = (int)(100 * soundSizeDone / soundSizeAll);
                        if (p != percent) {
                            percent = p;
                            ProgressCallbackThread(percent);
                        }
                    }
                    if (tag.m_record.m_version < TTagVersion.V_2_2)
                        bw.Write(tagData);
                    if (tag.m_record.m_version < TTagVersion.V_2_2)
                        tag.m_soundStartPos = 0;
                    else
                        tag.m_soundStartPos = tagData.Length;
                    bw.Close();
                    bsOut.Close();
                    fsOut.Close();
                    br.Close();
                    fsIn.Close();
                }
            }

            for (int index = 0; index < count; index++) {
                TTagItem tag = m_tags[index];
                if (tag.m_record.m_modify) {
                    if (bSaveTemp[index])
                        File.Delete(tag.m_filePath);
                }
            }

            for (int index = 0; index < count; index++) {
                TTagItem tag = m_tags[index];
                if (tag.m_record.m_modify) {
                    string path = Path.Combine(Path.GetDirectoryName(tag.m_filePath), tag.m_record.m_fileName);
                    if (bSaveTemp[index]) {
                        try {
                            File.Move(tag.m_tempFilePath, path);
                        }
                        catch (Exception e) {
                        }
                    }
                    tag.m_filePath = path;
                    //////////////////////////////////////////
                    // t@CXVێׂǂB
                    // 
                }
                tag.m_record.m_loadEnc = tag.m_record.m_saveEnc;
                tag.m_record.m_modify = false;

                // \̍XVB
                //if (MainViewRedrawCallback != null) {
                //    List<TTagItem> tags = new List<TTagItem>();
                //    tags.Add(tag);
                //    MainViewRedrawCallbackThread(tags);
                //}
            }
            MainViewRedrawCallbackThread(m_tags);

            m_work = null;
            SaveEndCallbackThread();
        }

        public bool Export(string filePath, string ext) {
            FileStream fs = new FileStream(filePath, FileMode.Create, FileAccess.Write);
            if (fs == null)
                return false;
            StreamWriter sw = new StreamWriter(fs);
            foreach (TTagItem t in m_tags) {
                string s = "";
                if (ext.Equals(".tsv"))
                    s = t.m_record.GetTSVLine();
                else if (ext.Equals(".csv"))
                    s = t.m_record.GetCSVLine();
                sw.WriteLine(s);
            }
            sw.Close();
            fs.Close();
            return true;
        }

        public bool Import(string filePath, string ext) {
            FileStream fs = new FileStream(filePath, FileMode.Open, FileAccess.Read);
            if (fs == null)
                return false;
            StreamReader sr = new StreamReader(fs);
            List<string> lines = new List<string>();
            while (true) {
                string s = sr.ReadLine();
                if (s == null || s.Length == 0)
                    break;
                lines.Add(s);
            }
            sr.Close();
            fs.Close();
            for (int i = 0; i < m_tags.Count && i < lines.Count; i++) {
                if (ext.Equals(".tsv"))
                    m_tags[i].m_record.SetTSVLine(lines[i]);
                else if (ext.Equals(".csv"))
                    m_tags[i].m_record.SetCSVLine(lines[i]);
            }
            return true;
        }

        public bool IsModified() {
            foreach (TTagItem t in m_tags) {
                if (t.m_record.m_modify)
                    return true;
            }
            return false;
        }

        void ProgressCallbackThread(int percent) {
            if (m_callbackTargetControl != null && ProgressCallback != null)
                m_callbackTargetControl.Invoke(ProgressCallback, percent);
        }

        void SaveEndCallbackThread() {
            if (m_callbackTargetControl != null && SaveEndCallback != null)
                m_callbackTargetControl.Invoke(SaveEndCallback);
        }

        void MainViewRedrawCallbackThread(List<TTagItem> tags) {
            if (m_callbackTargetControl != null && MainViewRedrawCallback != null)
                m_callbackTargetControl.Invoke(MainViewRedrawCallback, tags);
        }

    } // class
} // namespace
